# -*- coding: utf-8 -*-
"""
STILT stations by nature have only a geolocation [lat, lon], but no
other geographical information attached. This helper script
runs through all the available stilt stations, to
1. reverse geocode the location 
2. store all the information into a json file for offline retrieval.
(the json file will be distributed with the library for the moment)

The reason behind this, we cannot afford  the time to reverse geocode
all the stations 'live'. However the script can only run on our server
where we have access to the stilt station lat/lon information, this is
currently not available through http. ideally it will be provided through
https://stilt.icos-cp.eu/viewer/stationinfo

"""

import json
from icoscp.station import station as cpstation
import icoscp.const as CPC
import os
import pandas as pd
from tqdm import tqdm
import numpy as np
from icoscp import country


STN = 'stations.json'

def _save():
    ''' please DO NOT execute this function
        unless you want to re-create / update the offline station
        list, stored inside this script. If this is the case, 
		please execute this script on a CarbonPortal server, 
		with access to the stilt data file system.
		The content of the resulting stations.json file should be
		copied into this file to replace the current directory
    '''
        
    with open(STN, 'w') as f:
        json.dump(__save_all(), f)
        
def get(id=None):
    """
    Returns all offline STILT stations geographica info.

    Parameters
    ----------
    id : STR, default None -> which returns ALL stations
        Stilt station ID

    Returns
    -------
    DICT
        Geographical information locally stored for the Stilt station.
        If lat/lon is within a countryborder, the result
        from icoscp.coutntry9[lat, lon]) is returned.

    """
    
    d = os.path.abspath(__file__)
    stn = os.path.join(os.path.split(d)[0], 'stations.json')
    with open(stn, 'r') as f:
        data = json.loads(f.read())
    
    if id in data.keys():
        return data[id]
    
    return data


def __save_all():
    """ get all stilt stations available on the server 
        return dictionary with meta data, keys are stilt station id's
    """
        
    # use directory listing from siltweb data
    allStations = os.listdir(CPC.STILTPATH)

    # add information on station name (and new STILT station id)
    # from stations.csv file used in stiltweb.
    # this is available from the backend through a url
    df = pd.read_csv(CPC.STILTINFO)
    
    # add ICOS flag to the station
    icosStations = cpstation.getIdList()
    icosStations = list(icosStations['id'][icosStations.theme=='AS'])
    
    # dictionary to return
    stations = {}

    # fill dictionary with ICOS station id, latitude, longitude and altitude
    for ist in tqdm(sorted(allStations)):
  
        if not ist in df['STILT id'].values:
            continue
        
        stations[ist] = {}
        # get filename of link (original stiltweb directory structure) and extract location information
       
        loc_ident = os.readlink(CPC.STILTPATH+ist)
        clon = loc_ident[-13:-6]
        lon = np.float(clon[:-1])
        if clon[-1:] == 'W':
            lon = -lon
        clat = loc_ident[-20:-14]
        lat = np.float(clat[:-1])
        if clat[-1:] == 'S':
            lat = -lat
        alt = np.int(loc_ident[-5:])

        stations[ist]['lat']=lat
        stations[ist]['lon']=lon
        stations[ist]['alt']=alt
        stations[ist]['locIdent']=os.path.split(loc_ident)[-1]

    for s in tqdm(stations):
        stations[s]['geoinfo'] = country.get(latlon=[stations[s]['lat'],stations[s]['lon']])
        
    return stations

def __stationName(idx, name, alt):    
    if name=='nan':
        name = idx
    if not (name[-1]=='m' and name[-2].isdigit()):           
        name = name + ' ' + str(alt) + 'm'    
    return name



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    Define global (to icoscp) common tool to search for
    country information based on a static local file
    country.json  -> credit to https://github.com/mledoze/countries
    and for reverse geocoding -> credit to
    https://nominatim.openstreetmap.org
"""

import importlib.resources as pkgres
import json
import requests
import icoscp

def get(**kwargs):
    """
    Search country information.
    Please note: in case you provide more than one parameter, the order
    of keywords is not respected. The execution order is always like
    the function signature and as soon as a result is found, it will be
    returned and the search is stopped.

    Accepted keywords: code='', name='', latlon=[], search=''

    Example:
        .get()                      list of dict: all countries
        .get(code='CH')             dict: Switzerland
        .get(name='greece')         dict: Greece
        .get(latlon=[48.85, 2.35])  dict:
        .get(search='europe')

    Parameters
    ----------
    code : STR
        Search by ISO 3166-1 2-letter or 3-letter country codes

    name : STR
        search by country name, including alternativ spellings.
        It can be the native name or a partial name.

    latlon : List[]
        List with two integer or floating point numbers representing
        latitude and longitude. BE AWARE: using an external service
        from openstreetmap for reverse geocoding

    search : STR
        arbitrary text search, not case sensitiv, search in all fields

    Returns
    -------
    DICT: if a single country is found
    LIST[DICT]: list of dicts if more than one countre
    BOOL (False) if no result

    """

    # create a ressource file and read
    countries = pkgres.read_text(icoscp, 'countries.json')
    countries  = json.loads(countries)

    if not kwargs:
        return countries

    if 'code' in kwargs.keys():
        return _c_code(kwargs['code'], countries)


    if 'name' in kwargs.keys():
        return _c_name(kwargs['name'], countries)

    if 'search' in kwargs.keys():
        return _c_search(kwargs['search'], countries)

    if 'latlon' in kwargs.keys():
        latlon = kwargs['latlon']
        if isinstance(latlon, list) and len(latlon) == 2:
            country = _c_reverse(latlon)
        if country:
            return _c_code(country, countries)

    return False

def _c_search(search, countries):
    country = []
    for ctn in countries:
        if search.lower() in str(ctn).lower():
            country.append(ctn)

    if not country:
        return False

    if len(country) ==1 :
        #return the dictionary rather than the list
        return country[0]

    return country


def _c_code(code, countries):
    country = []
    for ctn in countries:
        if code.lower() == str(ctn['cca2']).lower() or \
            code.lower() == str(ctn['cca3']).lower():

            country.append(ctn)
    if not country:
        return False

    if len(country) ==1 :
        #return the dictionary rather than the list
        return country[0]

    return country

def _c_name(name, countries):
    country = []
    for ctn in countries:
        if name.lower() in str(ctn['name']).lower() \
             or name.lower() in str(ctn['altSpellings']).lower():

            country.append(ctn)


    if not country:
        return False

    if len(country) ==1 :
        #return the dictionary rather than the list
        return country[0]

    return country

def _c_reverse(latlon):
    # revers geocoding
    base = 'https://nominatim.openstreetmap.org/reverse?format=json&'
    url = base + 'lat=' + str(latlon[0]) + '&lon=' + str(latlon[1]) + '&zoom=3'
    response = requests.get(url=url)
    country = response.json()
    if 'address' in country.keys():
        return country['address']['country_code']

    return False


if __name__ == "__main__":

    MSG = """

    # find country information from a static file with the icoscp library.
    # you can import this file with:

    from icoscp import country

    # arbitrary text search
    a = get(search='Europe')

    # search by country code (alpha2 & alpha3)
    b = get(code='SE')      # returns Sweden
    c = get(code='CHE')     # returns Switzerland

    # search by name, includes alternative spellings
    d = get(name = 'greece') # returns Greece
    e = get(name = 'helle' ) # returns Greece and Seychelles

    # search by lat lon...!! BE AWARE this is using an external
    # rest API from OpenStreeMap https://nominatim.openstreetmap.org
    f = get(latlon=[42.5,13.8]) # returns Italy
    """
    print(MSG)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    Extract all STILT stations from the ICOS Carbon Portal Server
    The main function is
    station.find() to search and filter STILT stations and
    station.get() to create a STILT station object with access to the data.

"""

__credits__     = "ICOS Carbon Portal"
__license__     = "GPL-3.0"
__version__     = "0.1.0"
__maintainer__  = "ICOS Carbon Portal, elaborated products team"
__email__       = ['info@icos-cp.eu', 'claudio.donofrio@nateko.lu.se']
__status__      = "rc1"
__date__        = "2021-09-27"

import os
import json
import numpy as np
import pandas as pd

from tqdm.notebook import tqdm

from icoscp.station import station as cpstation
from icoscp.stilt.stiltobj import StiltStation
from icoscp.stilt import geoinfo
from icoscp.stilt import fmap
import icoscp.const as CPC
import icoscp.country

from icoscp.stilt import timefuncs as tf

# --- START KEYWORD FUNCTIONS ---

def _id(kwargs, stations):

    ids = kwargs['id']
    if isinstance(ids,str):
        ids=[ids]
    if not isinstance(ids,list):
        ids=list(ids)

    # to make the search not case sensitive we need to convert all
    # id's to CAPITAL LETTERS. The stilt on demand calculator
    # allows only captial letters.

    ids = [id.upper() for id in ids]

    # check stilt id
    flt = list(set(ids).intersection(stations))

    #check icos id
    for k in stations:
        if stations[k]['icos']:
            if stations[k]['icos']['stationId'] in ids:
                flt.append(k)

    stations =  {k: stations[k] for k in flt}
    return stations

def _outfmt(kwargs, stations):
    '''
    Parameters
    ----------
    kwargs : STR
        Define the output format. by default a 'DICT' is returned
        Possible arguments are:   
            - dict (default), with Station ID as key
            - pandas
            - list , provide a list of stilstation objects, equivalent to 
                    station.get([list of ids])
            - map (folium)
    '''
    
    if not stations:
        # no search result, return empty
        stations={'empty':'no stiltstations found'}
        return stations

    if 'outfmt' in kwargs:        
        fmt = kwargs['outfmt'].lower()
    else:
        fmt = 'dict'
    
    # make sure we have a valid keyword, else set default dict
    valid = ['dict', 'list', 'pandas', 'map']
    if not fmt in valid:
            fmt = 'dict'
    
    if fmt == 'dict':        
        #default
        return stations
    
    if fmt == 'pandas':
        df = pd.DataFrame().from_dict(stations)
        return df.transpose()

    if fmt == 'list':
        stnlist = __get_object(stations)
        return stnlist

    if fmt == 'map':
        return fmap.get(stations)

def _country(kwargs, stations):
    """
    Find all stations belonging to a country based on information from
    alpha2Code', 'alpha3Code', 'name', 'nativeName', 'altSpellings', 'demonym'
    Be aware, that STILT stations can have arbitrary lat/lon which
    are in international waters and hence not associated with a country

    """
    countries = kwargs['country']
    if isinstance(countries,str):
        countries=[countries]
    if not isinstance(countries,list):
        countries=list(countries)

    idx = []
    for k in stations:
        if stations[k]['geoinfo'] is False:
            # if there is no geoinfo, skip
            continue

        for c in countries:
            # the search keys are a combination of keys from the online
            # search at restcountries.eu and the static country file
            if len(c) < 4:
                searchKeys = {'alpha2Code', 'alpha3Code', 'cca2', 'cca3','ccn3', 'cioc'}
            else:
                searchKeys = {'name', 'nativeName', 'altSpellings', 'demonym', 'demonyms'}
            
            g = stations[k]['geoinfo']

            # make sure we have only valid search keys...add two sets
            for sk in searchKeys & g.keys():
                if c.lower() in str(g[sk]).lower():
                    idx.append(k)
                    break

    flt = list(set(idx).intersection(stations))
    stations =  {k: stations[k] for k in flt}
    return stations


def _bbox(kwargs, stations):
    """
    Find all stations within a lat/lon bounding box. Expected keyword argument
    input is bbox=[Topleft NorthWest (lat,lon),
                   BottomRight South East(lat,lon)]

    """
    bbox=kwargs['bbox']
    lat1 = float(bbox[1][0])
    lat2 = float(bbox[0][0])
    lon1 = float(bbox[0][1])
    lon2 = float(bbox[1][1])

    flt = []
    for k in stations:
        if lat1 <= float(stations[k]['lat']) <= lat2:
            if lon1 <= float(stations[k]['lon']) <= lon2:
                flt.append(k)

    stations =  {k: stations[k] for k in flt}
    return stations

def _pinpoint(kwargs, stations):
    # the user MUST provide lat, lon, but only optional
    # the distance in KM to create the boundding box.
    # set default 200km bounding box..
    if len(kwargs['pinpoint']) == 2:
        kwargs['pinpoint'].append(200)
    lat = kwargs['pinpoint'][0]
    lon = kwargs['pinpoint'][1]
    deg = kwargs['pinpoint'][2]*0.01
    lat1 = lat+deg
    lat2 = lat-deg
    lon1 = lon-deg
    lon2 = lon+deg
    box = {'bbox':[(lat1, lon1),(lat2, lon2)]}
    return _bbox(box, stations)

def _dates(kwargs, stations):
    """
    Check availability for specific dates.
    input needs to be a list. Stations are returned which have data
    for any dates in the list. Remember, only year and month is checked
    NOT single days.
    """

    # return empyt if dates is not a list
    if not isinstance(kwargs['dates'],list):
        print('Dates must be a list')
        return {}

    # parse all dates to a clean list
    dates = [tf.parse(d) for d in kwargs['dates']]
    # remove Nones
    dates = [d for d in dates if d]
    if not dates:
        return {}

    flt = []
    for st in stations:
        # check each station for all dates...
        # use daterange with sdate and edate set to the same.
        for d in dates:
            if tf.check_daterange(d, d, stations[st]):
                flt.append(st)

    stations =  {k: stations[k] for k in flt}
    return stations

def __daterange(kwargs, stations):
    """
    Check availability for a date range.
    sdate AND edate is provided in the arguments. Daterange is PRIVATE
    function...and cant be called directly, but is called if
    sdate AND edate is provided in the arguments
    """

    sdate = tf.parse(kwargs['daterange'][0])
    edate =  tf.parse(kwargs['daterange'][1])

    # return an empyt dict, if date is not a date object
    if not sdate or not edate:
        return {}

    if edate < sdate:
        print('Start-date is set to a later date than end-date... ')
        return {}

    flt = []
    for st in stations:
        if tf.check_daterange(sdate, edate, stations[st]):
            flt.append(st)
    stations =  {k: stations[k] for k in flt}
    return stations



def _sdate(kwargs, stations):
    """
    Find all stations with valid data for >= sdate (year and month only)
    """
    #Convert date-string to date obj:
    sdate =  tf.parse(kwargs['sdate'])

    # return and empyt dict, if sdate is not a date object
    if not sdate:
        print("Check date format")
        return {}

    flt = []
    for st in stations:
        if tf.check_smonth(sdate, stations[st]):
           flt.append(st)

    stations =  {k: stations[k] for k in flt}
    return stations


def _edate(kwargs, stations):
    """
    Find all stations with valid data for <= edate (year and month only)
    """
    edate =  tf.parse(kwargs['edate'])

    # return an empyt dict, if sdate is not a date object
    if not edate:
        print("Check date format")
        return {}

    flt = []
    for st in stations:
        if tf.check_emonth(edate, stations[st]):
           flt.append(st)

    stations =  {k: stations[k] for k in flt}
    return stations


def _search(kwargs, stations):
    """ search for arbitrary string"""
    idx = []
    for k in stations:
        txt = json.dumps(stations[k])
        if kwargs['search'].lower() in txt.lower():
            idx.append(k)

    flt = list(set(idx).intersection(stations))
    return {k: stations[k] for k in flt}

# --- END KEYWORD FUNCTIONS ---


def __get_object(stations):

    return [StiltStation().get_info(stations[st]) for st in stations.keys()]


def __get_all():
    """ get all stilt stations available on the server
        return dictionary with meta data, keys are stilt station id's
    """

    # use directory listing from siltweb data
    allStations = os.listdir(CPC.STILTPATH)

    # add information on station name (and new STILT station id)
    # from stations.csv file used in stiltweb.
    # this is available from the backend through a url
    df = pd.read_csv(CPC.STILTINFO)

    # add ICOS flag to the station
    icosStations = cpstation.getIdList()
    icosStations = list(icosStations['id'][icosStations.theme=='AS'])

    # dictionary to return
    stations = {}

    # fill dictionary with ICOS station id, latitude, longitude and altitude
    for ist in tqdm(sorted(allStations)):
        if not ist in df['STILT id'].values:
            continue

        stations[ist] = {}
        # get filename of link (original stiltweb directory structure) and
        # extract location information

        loc_ident = os.readlink(CPC.STILTPATH+ist)
        clon = loc_ident[-13:-6]
        lon = np.float(clon[:-1])
        if clon[-1:] == 'W':
            lon = -lon
        clat = loc_ident[-20:-14]
        lat = np.float(clat[:-1])
        if clat[-1:] == 'S':
            lat = -lat
        alt = np.int(loc_ident[-5:])

        stations[ist]['lat']=lat
        stations[ist]['lon']=lon
        stations[ist]['alt']=alt
        stations[ist]['locIdent']=os.path.split(loc_ident)[-1]

        # set the name and id
        stations[ist]['id'] = df.loc[df['STILT id'] == ist]['STILT id'].item()
        stationName = str(df.loc[df['STILT id'] == ist]['STILT name'].item())
        stations[ist]['name'] = __stationName(stations[ist]['id'], stationName, stations[ist]['alt'])

        # set a flag if it is an ICOS station
        stn = stations[ist]['id'][0:3].upper()
        if stn in icosStations:
            stations[ist]['icos'] = cpstation.get(stn).info()
        else:
            stations[ist]['icos'] = False

        # set years and month of available data
        years = os.listdir(CPC.STILTPATH+'/'+ist)
        stations[ist]['years'] = years
        for yy in sorted(stations[ist]['years']):
            stations[ist][yy] = {}
            months = os.listdir(CPC.STILTPATH+'/'+ist+'/'+yy)
            # remove cache txt entry
            sub = 'cache'
            months = sorted([m for m in months if not sub.lower() in m.lower()])
            stations[ist][yy]['months'] = months
            stations[ist][yy]['nmonths'] = len(stations[ist][yy]['months'])

    # merge geoinfo
    geo = geoinfo.get()

    for k in stations:
        if k in geo.keys():
            stations[k]['geoinfo'] = geo[k]['geoinfo']
        else:
            stations[k]['geoinfo'] = __country([stations[k]['lat'],stations[k]['lon']])

    return stations


def __country(latlon):
    # be aware, that by providing latlon
    # triggers an external reverse geocoder.
    return icoscp.country.get(latlon=latlon)


def __stationName(idx, name, alt):
    if name=='nan':
        name = idx
    if not (name[-1]=='m' and name[-2].isdigit()):
        name = name + ' ' + str(alt) + 'm'
    return name


def find(**kwargs):
    """
    Return a list of stilt stations. Providing no keyword arguments will
    return a complete list of all stations. You can filter the result by
    by providing the following keywords:

    Parameters
    ----------

    id STR, list of STR:
        Provide station id or list of id's. You can provide
        either stilt or icos id's mixed together.
        Example:    station.find(id='HTM')
                    station.find(id=['NOR', 'GAT344'])

    search STR:
        Arbitrary string search keyword
         Example:    station.find(search='north')

    stations DICT
        all actions are performed on this dictionary, rather than
        dynamically search for all stilt station on our server.
        Can be useful for creating a map, or getting a subset of stations
        from an existing search.
        Example: myStations = station.find(search='north')
                 refined = station.find(stations=myStations, country='Finland')

    # Spatial search keywords:

    country STR, list of STR:
        Provide country code either fullname, 2 or 3 digit
        where country is e.g "SE", "SWE" or "Sweden"
        Example:    station.find(country='Sweden')

    bbox LIST of Tuples:
        spatial filter by bounding box where
        bbox=[Topleft(lat,lon), BottomRight(lat,lon)]
        The following example is approximately covering scandinavia
        Example:    station.find(bbox=[(70,5),(55,32)])

    pinpoint LIST: [lat, lon, distanceKM]
        spatial filter by pinpoint location plus distance in KM
        distance in km is use to create a bounding box
        We use a very rough estimate of 1degree ~ 100 km
        Example: stations.find(pinpoint = [40.42, -3.70, 500]
    # Temporal search keywords:
    Be aware, that the granularity for all temporal keywords is year and month
    (days are not considered in the search): input format for the dates entry
    MUST be convertible to data time object throug pandas.
    -> pandas.to_datetime(date)

    sdate Input formats:
        - datetime.date objs
        - FLOAT or INT unix timestamp
        - pandas.datetime
        - STR: "YYYY-MM-DD" , "YYYY", "YYYY/MM/DD":
        where 'sdate' (Start Date) is a single entry. If you provide ONLY
        sdate, stations with any data available for >= sdate is returned
    Example: station.find(sdate= '2018-05-01')

    edate Input format see sdata:
        - datetime.date objs
        - FLOAT or INT unix timestamp
        - pandas.datetime
        - STR: "YYYY-MM-DD" , "YYYY", "YYYY/MM/DD":
        where 'edate' (End Date) is a single entry. If you provide ONLY
        edate, stations with any data available  <= edate is returned
    Example: station.find(edate='2018-06-01')

    If you provide sdate AND edate, any station with available data
    within that date range is retured.

    dates [] is a list of dates.
        This will return a list of stations where data is available for
        for any of the provided dates. Input format, see sdate,edate.
        Remember, that only year and month is checked.
    Example: station.find(dates=['2020-01-01', '2020/05/23'])

    outfmt STR ['dict'| 'pandas' | 'list' | 'map']:
        the result is returned as
        
            dict:       dictionary with full metadata for each station
                        Stiltstation ID is the 'key' for the dict entries
            pandas:     dataframe with some key information    
            list:       list of stilt station objects
            map:        folium map, can be displayed directly in notebooks
                        or save to a static (leaflet) .html webpage

    Returns
    -------
    List of Stiltstation in the form of outfmt=format, see above.
    Default DICT

    """

    if 'stations' in kwargs:
        stations = kwargs['stations']
    else:
        # start with getting all stations
        stations = __get_all()

    # with no keyword arguments, return all stations
    # in default format (see _outfmt())
    if not kwargs:
        return _outfmt(kwargs, stations)

    # convert all keywords to lower case
    kwargs =  {k.lower(): v for k, v in kwargs.items()}


    # check if sdate AND edate is provided. If yes,
    # create a date_range entry and remove sdate and edate:
    if 'sdate' in kwargs.keys() and 'edate' in kwargs.keys():
        kwargs['daterange'] = [kwargs['sdate'], kwargs['edate']]
        del kwargs['sdate']
        del kwargs['edate']



    # valid key words. Make sure all are lower capital and that
    # the function has been defined above and that outfmt is the very
    # last call:
    fun = {'id': _id,
           'country': _country,
           'bbox': _bbox,
           'pinpoint': _pinpoint,
           'sdate':_sdate,
           'edate':_edate,
           'dates':_dates,
           'daterange':__daterange,
           'search':_search}

    for k in kwargs.keys():
        if k in fun.keys():
            stations = fun[k](kwargs, stations)

    return _outfmt(kwargs, stations)

def get(id=None):
    """
    This function returns a stilt station object or a list of
    stiltstation objects.
    A stilt station object, gives access to the underlying data
    (timeseries and footprints)
    You may provide a str or list of STILT id's or the 'result' of a .find()

    Example: .get('HTM030')
            .get(['HTM030'])
                returns one stiltobject for HTM030

             .get(['HTM030','HTM150'])
                 returns a list of two stiltobjects

            .get('KIT')
                returns a list of stiltstations based on .find(search='KIT')

            .get(find(country=['Sweden','Finland']))
                returns a list of stilstations found in Sweden and Finland
    Parameters
    ----------
    id :    DICT | LIST[DICT]
                The result of .find(....) where one station (DICT) is found or
                multiple stations (LIST[DICT])

            STR | LIST[STR]
                A single string or a list of strings containing one or more
                stilt station ids.

    Returns
    -------
    LIST[stiltstation]

    """

    stationslist = []

    if isinstance(id,str):
        st = find(id=id)
        if not st:
            return None
        for s in st:
            stationslist.append(StiltStation(st[s]))

    if isinstance(id,dict):
        for s in id:
            stationslist.append(StiltStation(id[s]))

    if isinstance(id, list):
        if isinstance(id[0],dict):
           for s in id:
               stationslist.append(StiltStation(id[s]))

        if isinstance(id[0], str):
            st = find(id=id)
            if not st:
                return None
            for s in st:
                stationslist.append(StiltStation(st[s]))

    if len(stationslist) == 1:
        return stationslist[0]
    else:
        return stationslist




#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""    
    create a map based on folium (wrapper for leaflet)
"""

__credits__     = "ICOS Carbon Portal"
__license__     = "GPL-3.0"
__version__     = "0.1.0"
__maintainer__  = "ICOS Carbon Portal, elaborated products team"
__email__       = ['info@icos-cp.eu', 'claudio.donofrio@nateko.lu.se']
__status__      = "rc1"
__date__        = "2021-04-16"


import folium
from folium.plugins import MarkerCluster


def get(stations, fmt='map', cluster=True):
    '''
    Provide a map based on the folium module.

    Parameters
    ----------
    stations : DICT : StiltStation dictionary
    fmt: STR: ['map' | 'html'] see Returns below.
    cluster: BOOL: default True. Markers are clustered or not

    Returns
    -------
    Depending on fmt:
        By default a folium map object is returned,
        which can be displayed directly in a Jupyter Notebook
        fmt='html' returns a static html website as STR which can
        be saved to a file.
    '''
    
    if not isinstance(stations, dict):
        return 'input needs to be dictionary'
    
    zoom = 4

    #create the map
    myMap = folium.Map(zoom_start=zoom,no_wrap=True)
    
    # add tiles, to see what kind of basemap you like
    # tile layers will be displayed in the top right
    
    folium.TileLayer('openstreetmap').add_to(myMap)
    folium.TileLayer('cartodbpositron').add_to(myMap)
    folium.TileLayer('stamentoner').add_to(myMap)
    folium.TileLayer('stamenterrain').add_to(myMap)
    
    markers = []    # keep all the markers for clustering
    lats = []       # keep lat, lon vector to calculate the center of the map
    lons = []       # keep lat, lon vector to calculate the center of the map
    for k in stations:
        lat = float(stations[k]['lat'])
        lats.append(lat)
        lon = float(stations[k]['lon'])
        lons.append(lon)
        msg = _pretty_html(stations[k])
        markers.append(folium.Marker(location=[lat, lon], popup=msg))
    
    if cluster:
        mc = MarkerCluster(name='StiltStations')
        for m in markers:
            mc.add_child(m)
        myMap.add_child(mc)
    else:
        for m in markers:
            myMap.add_child(m)

    #re-centre the map
    bounds = [(min(lats), min(lons)), (max(lats), max(lons))]
    folium.FitBounds(bounds).add_to(myMap)
    #myMap.fit_bounds(sw,ne)

    folium.LayerControl().add_to(myMap)    
    return myMap

def _pretty_html(station):
    # create a html table for the popup 
    header = """<!DOCTYPE html>
                <html>
                <link rel="stylesheet" href="https://www.w3schools.com/w3css/4/w3.css">
                <body><div class="w3-container">
                    <table width="200" class="w3-striped">
                """
    body = '<tr><td colspan="2">Stilt Station info: </td></tr>'
    body += '<tr><td>Stilt id: </td><td>' + station['id'] + '</td></tr>'
    if station['icos']:
        body += '<tr><td>ICOS id: </td><td><a href= "'
        body += station['icos']['uri'][0] + '">'
        body += station['icos']['stationId'] + "</a>" +'</td></tr>'
        
    body += '<tr><td>lat: </td><td>' + str(station['lat']) + '</td></tr>'
    body += '<tr><td>lon: </td><td>' + str(station['lon']) + '</td></tr>'    
    body += '<tr><td colspan="2">\
        <a href="https://stilt.icos-cp.eu/viewer/" target="_blank">\
        https://stilt.icos-cp.eu/viewer/</a></td></tr>'    
    footer = "</table></div></body></html>"
    html = header + body + footer
        
    return html



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""    
    this file is intended to provide a folium map object
    for a station selection. Please do not use directly, 
    but import station and change the output format.
    
    from icoscp.station import station
    station.getIdList(.......outfmt='map')
"""

__author__ = ["Zois Zogopoulos"]
__credits__ = "ICOS Carbon Portal"
__license__ = "GPL-3.0"
__version__ = "0.1.0"
__maintainer__ = "ICOS Carbon Portal, elaborated products team"
__email__ = ['info@icos-cp.eu', 'zois.zogopoulos@nateko.lu.se']
__status__ = "rc1"
__date__ = "2021-09-20"

# Standard library imports.
import json
# Related third party imports.
import folium
import pandas as pd
from folium.plugins import MarkerCluster
import requests


def get(queried_stations, project):
    """Generates a folium map of stations.

    Uses the requested stations dataframe along with the REST countries
    API to generate an interactive folium map. Each marker in the
    folium map represents a station of observations.

    Parameters
    ----------
    queried_stations : pandas.Dataframe
        `queried_stations` dataframe includes each station's landing
        page or (uri), id, name, country, lat, lon, elevation, project,
        and theme.

    project : str
        The name of the project that the user inserted in the caller
        function in order to search for stations.

    Returns
    -------
    stations_map : folium.Map
        `stations_map` is an interactive map used for visualizing
        geospatial data.

    """
    # Request countries data online.
    response = request_rest_countries()
    # Edit the requested data.
    edited_response = collect_rest_data(response)
    # Apply countries data on the queried stations and remove stations
    # without a fixed location.
    stations = edit_queried_stations(queried_stations, edited_response)
    stations_map = folium.Map()
    # Provide the map name within the top right menu.
    if project == 'ALL':
        cluster_name = ', '.join(['ICOS', 'NEON', 'INGOS', 'FLUXNET'])
    else:
        cluster_name = project
    marker_cluster = MarkerCluster(name=cluster_name)
    # Add tile layers to the folium map. Default is 'openstreetmap'.
    add_tile_layers(stations_map)
    # Use the stations at the most southwest and northeast locations
    # and bind the map within these stations.
    sw_loc = stations[['lat', 'lon']].dropna(axis=0).min().values.tolist()
    ne_loc = stations[['lat', 'lon']].dropna(axis=0).max().values.tolist()
    stations_map.fit_bounds([sw_loc, ne_loc])
    stations = stations.transpose()
    for station_index in stations:
        # Collect each station's info.
        station_info = stations[station_index]
        # Create the html popup message for each station.
        popup = folium.Popup(generate_popup_html(station_info, response))
        if response['service']:
            # Set the icon for each marker using the country's flag.
            icon = folium.CustomIcon(icon_image=station_info.flag, icon_size=(20, 14))
        else:
            icon = folium.Icon(color='blue', icon_color='white', icon='info_sign')
        # Add a marker for each station at the station's location
        # along with the popup and the tooltip.
        station_marker = folium.Marker(location=[station_info.lat, station_info.lon],
                                       tooltip='<b>' + station_info.id + '</b>',
                                       popup=popup,
                                       icon=icon)
        # Add the station marker to the cluster.
        marker_cluster.add_child(station_marker)
    # Add the cluster and the layer control to the folium map.
    stations_map.add_child(marker_cluster)
    stations_map.add_child(folium.LayerControl())
    return stations_map


def request_rest_countries():
    """Requests data from rest-countries API.

    This function uses https://restcountries.com/ API to request data
    (names ,country-codes and flags) for countries.

    Returns
    -------
    response : dict
        Returns a dictionary with countries data obtained from
        rest-countries API. The `service` key validates the source of
        the data ('com', or False). If the request fails the `service`
        key has a value of False.

    Raises
    ------
    HTTPError
        An HTTPError exception is raised if the requested REST
        countries data is unavailable.
    SSLError
        An SSLError exception is raised if the requested resources have
        an untrusted SSL certificate.

    """

    response = {'service': False}
    response_com = None
    try:
        # Try to request countries data from
        # https://restcountries.com/ REST-ful API.
        response_com = requests.get(
            'https://restcountries.com/v2/all?fields=name,flags,alpha2Code')
        response_com.raise_for_status()
    except (requests.exceptions.HTTPError, requests.exceptions.SSLError) as e:
        print('Restcountries \'.com\' request error: ' + str(e))

    if response_com:
        response = {'service': 'com', 'data': response_com}
    # If the request failed, the response will contain a False service
    # value which is then used by the caller function to generate the
    # folium map without any errors but with less information.
    return response


def collect_rest_data(response):
    """Extracts raw rest-countries data from requested resources.

    Parameters
    ----------
    response : dict
        If the requested resources were available the `response`
        dictionary will contain the raw rest-countries data.

    Returns
    -------
    response : dict
        Returns an updated version of the `response` dictionary
        obtained from `request_rest_countries()` function. If the
        request was successful, this version will include the
        rest-countries data that was extracted from the request.

    """

    # Rest-countries resources are available.
    if response['service']:
        json_countries = json.loads(response['data'].text)
        countries_data = {}
        # Use the requested data to create a dictionary of country
        # names, codes, and flags.
        for country in json_countries:
            code = country['alpha2Code']
            country_name = country['name']
            country_flag = country['flags']['svg']
            countries_data[code] = {'name': country_name, 'flag': country_flag}
        # Include the 'UK' alpha2code which is missing from
        # restcountries API.
        countries_data['UK'] = countries_data['GB']
        # Add the created dictionary to the response.
        response['countries_data'] = countries_data
    return response


def edit_queried_stations(queried_stations, edited_response):
    """Applies new data on queried stations.

    Uses the `edited_response` dictionary to add data to the queried
    stations which are then appended to a new dataframe.

    Parameters
    ----------
    queried_stations : pandas.Dataframe
        `queried_stations` dataframe includes each station's landing
        page or (uri), id, name, country, lat, lon, elevation, project,
        and theme.
    edited_response: dict
        `edited_response` is an updated version of the `response`
        dictionary obtained from `request_rest_countries()` function.
        If the request was successful, this version will include the
        rest-countries data that was extracted using the
        `collect_rest_data()` function.

    Returns
    -------
    edited_stations : pandas.Dataframe
        `edited_stations` is an updated version of the
        `queried_stations` dataframe which was obtained using a sparql
        query. Also this edited version is stripped off stations
        without a fixed position (Instrumented ships of opportunity).

    """

    edited_stations = pd.DataFrame()
    # Transpose the requested stations dataframe and iterate each
    # station.
    queried_stations = queried_stations.transpose()
    for station_index in queried_stations:
        # Collect each station's info from the sparql query.
        station_info = queried_stations[station_index]
        # Measurements collected from instrumented Ships of
        # Opportunity don't have a fixed location and thus are
        # excluded from the folium map.
        if station_info.lat is None or station_info.lon is None:
            continue
        # The requested resources are available.
        if edited_response['service']:
            countries_data = edited_response['countries_data']
            # Add new labels and data (using the rest-countries API) and
            # update existing labels of the dataframe to better represent
            # the station's information.
            station_info['country_code'] = station_info.pop('country')
            station_info['station_name'] = station_info.pop('name')
            station_info['country'] = countries_data[station_info.country_code]['name']
            station_info['flag'] = countries_data[station_info.country_code]['flag']
        edited_stations = edited_stations.append(other=station_info)
    return edited_stations


def add_tile_layers(folium_map):
    """Adds multiple layers to a folium map.

    Parameters
    ----------
    folium_map : folium.Map
        `folium_map` is the parent element to which all tile layers
        (children elements) will be added.

    Returns
    -------
    folium_map : folium.Map
        Returns an updated version of the `folium_map`.

    """

    # Add built-in tile layers.
    folium_map.add_child(folium.TileLayer('cartodbpositron'))
    folium_map.add_child(folium.TileLayer('cartodbdark_matter'))
    folium_map.add_child(folium.TileLayer('stamenwatercolor'))
    folium_map.add_child(folium.TileLayer('stamentoner'))
    folium_map.add_child(folium.TileLayer('stamenterrain'))
    # Add another layer with satellite images from ESRI.
    folium_map.add_child(folium.TileLayer(
        tiles='https://server.arcgisonline.com/ArcGIS/rest/services/World_Imagery/MapServer'
              '/tile/{z}/{y}/{x}',
        attr='Esri',
        name='Esri Satellite',
        overlay=False,
        opacity=1.0,
        control=True))
    return folium_map


def generate_popup_html(station_info, response):
    """Generates an html popup for an interactive folium map.

    For each station being processed creates an html popup with useful
    information.

    Parameters
    ----------
    station_info : pandas.Series
        `station_info` series contains an extended version of the
        station series in the `queried_stations` collection. This
        version includes a landing page or (uri), id, lat, lon,
        elevation, project, theme, country_code, station_name, country,
        and flag.
    response : dict
        An updated version of the `response` dictionary
        obtained from `request_rest_countries()` function. If any of
        HTTP requests were successful, this version will include the
        rest-countries data that was extracted from the requests.

    Returns
    -------
    folium_html : branca.Html
        `folium_html` contains all the necessary data to visualize a
        station's information as a popup within the folium map.

    """

    # Format station's html string using data extracted from the dataframe.
    if response['service']:
        station_html = \
            """
            <table border='0'>
                <caption style='font-weight:bold;font-size:18px;padding:15px;'>
                    Station Information
                </caption>
                <tr style='background-color:#f8f4f4'>
                    <td style="padding:4px"><nobr>Station name</nobr></td>
                    <td style="padding:4px"><nobr><a title="{uri}" href="{uri}">
                        {station_name}
                    </a></nobr></td>
                </tr>
                <tr><td style="padding:4px">Station ID</td><td style="padding:4px">{id}</td></tr>
                <tr style='background-color:#f8f4f4'>
                    <td style="padding:4px">Country, Country code</td>
                    <td style="padding:4px"><nobr>{country} - {country_code}</nobr></td>
                </tr>
                <tr>
                    <td style="padding:4px"><nobr>Latitude, Longitude, Elevation</nobr></td>
                    <td style="padding:4px"><nobr>{latitude}, {longitude}, {elevation}</nobr></td>
                </tr>
                <tr style='background-color:#f8f4f4'>
                    <td style="padding:4px">Project</td>
                    <td style="padding:4px">{project}</td>
                </tr>
                <tr><td style="padding:4px">Theme</td><td style="padding:4px">{theme}</td></tr>
            </table>
            """.format(uri=station_info.uri, station_name=station_info.station_name,
                       id=station_info.id, country=station_info.country,
                       country_code=station_info.country_code, latitude=station_info.lat,
                       longitude=station_info.lon, elevation=station_info.elevation,
                       project=station_info.project, theme=station_info.theme)
    else:
        station_html = \
            """
            <table border='0'>
                <caption style='font-weight:bold;font-size:18px;padding:15px;'>
                    Station Information
                </caption>
                <tr style='background-color:#f8f4f4'>
                    <td style="padding:4px"><nobr>Station name</nobr></td>
                    <td style="padding:4px"><nobr><a title="{uri}" href="{uri}">
                        {station_name}
                    </a></nobr></td>
                </tr>
                <tr><td style="padding:4px">Station ID</td><td style="padding:4px">{id}</td></tr>
                <tr style='background-color:#f8f4f4'>
                    <td style="padding:4px">Country code</td>
                    <td style="padding:4px"><nobr>{country_code}</nobr></td>
                </tr>
                <tr>
                    <td style="padding:4px"><nobr>Latitude, Longitude, Elevation</nobr></td>
                    <td style="padding:4px"><nobr>{latitude}, {longitude}, {elevation}</nobr></td>
                </tr>
                <tr style='background-color:#f8f4f4'>
                    <td style="padding:4px">Project</td>
                    <td style="padding:4px">{project}</td>
                </tr>
                <tr><td style="padding:4px">Theme</td><td style="padding:4px">{theme}</td></tr>
            </table>
            """.format(uri=station_info.uri, station_name=station_info['name'],
                       id=station_info.id, country_code=station_info.country,
                       latitude=station_info.lat, longitude=station_info.lon,
                       elevation=station_info.elevation, project=station_info.project,
                       theme=station_info.theme)
    # Render html from string.
    folium_html = folium.Html(station_html, script=True)
    return folium_html









#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    Created on Fri Aug  9 10:40:27 2019
    This file handles the communication with the ICOS Sparql Endpoint.
    Instances of RunSparql() is quering the ICOS Sparql endpoint, 
    and returns a formatted response.
"""

__author__      = ["Claudio D'Onofrio"]
__credits__     = "ICOS Carbon Portal"
__license__     = "GPL-3.0"
__version__     = "0.1.0"
__maintainer__  = "ICOS Carbon Portal, elaborated products team"
__email__       = ['info@icos-cp.eu', 'claudio.donofrio@nateko.lu.se']
__status__      = "rc1"
__date__        = "2019-08-09"


import requests
import pandas as pd

class RunSparql():
    """
        Class to send a sparql query to the icos endpoint and get
        formated output back.
        :param sparql_query, string, valid query
        :param output_format, define format of returned object ['json', 'csv', 'array', 'dict', 'pandas']
        :return False, if query is not successful otherwise output_format(results)
    """

    def __init__(self, sparql_query='', output_format='txt'):

        self.format = output_format
        self.query = sparql_query
        self.__result = False

    @property
    def format(self):
        """ contains the output format for the results from the ICOS sqparql endpoint"""
        return self.__format

    @format.setter
    def format(self, fmt):
        """
            You can set the format of the results returned after running the query.
            Allowed formats are:
            - 'json' (default), string
            - 'csv' a string object with comma separated values, where the the first row contains the column names
            Python specific formats:
            - 'dict' a python dict representing the json object
            - 'pandas' a pandas table with column names
            - 'list'
            - 'array' returns TWO python arrays, the first with the headers, the second with the values.
        """
        allowed = ['json', 'csv', 'dict', 'pandas', 'array', 'html']
        try:
            fmt = str(fmt)
        except TypeError:
            return 0

        if fmt.lower() in allowed:
            self.__format = fmt.lower()
        else:
            self.__format = 'json'

    @property
    def query(self):
        """ Import a sparql query. No validation is performed. """
        return self.__query

    @query.setter
    def query(self, query):

        try:
            query = str(query)
        except TypeError:
            return 0
        self.__query = str(query)

    # ------------------------------------------------------------------------
    def data(self):
        return self.__result

    # ------------------------------------------------------------------------
    def run(self):
        """
            This functions queries the ICOS sparql endpoint queryString.
            By default the returned object is the raw "json" object.
        """
        # check if query is set
        if not len(self.__query):
            print('no query found')
            return

        url = 'https://meta.icos-cp.eu/sparql'

        r = requests.get(url, params={'query': self.__query})
        if not r.ok:
            print(r.ok, r.reason)
            return r.ok, r.reason

        # now check what format we should return.
        # either set the __result directly, or call a transform
        # function which sets the __result.

        if self.format == 'json':
            self.__result = r.text
        if self.__format == 'dict':
            self.__result = r.json()
        if self.__format == 'csv':
            self.__result = self.__to_csv(r.json())
        if self.__format == 'pandas':
            self.__result = self.__to_pandas(r.json())
        if self.__format == 'array':
            self.__result = self.__to_array(r.json())
        if self.__format == 'html':
            self.__result = self.__to_html(r.json())
        return self.__result

    # ------------------------------------------------------------------------
    def __to_array(self, data):

        # convert the the result into two arrays
        # colName, colData
        colname = data['head']['vars']
        coldata = []

        for row in data['results']['bindings']:
            item = []
            for c in colname:
                item.append(row.get(c, {}).get('value'))
            coldata.append(item)
        return colname, coldata

    # ------------------------------------------------------------------------
    def __to_csv(self, data):
        colname, coldata = self.__to_array(data)

        # add the header line
        csvstring = ','.join(colname) + '\n'
        for row in coldata:
            csvstring += ','.join(row) + '\n'

        return csvstring

    # ------------------------------------------------------------------------
    def __to_pandas(self, data):

        colname, coldata = self.__to_array(data)

        return pd.DataFrame(coldata, columns=colname)
    # ------------------------------------------------------------------------

    def __to_html(self, data):
        colname, coldata = self.__to_array(data)

        return pd.DataFrame(coldata, columns=colname).to_html(classes='table table-striped table-hover')


















#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    Description:      Python functions that process time in STILT model outputs.    
"""

__author__      = ["Karolina Pantazatou", "Claudio DOnofrio"]
__credits__     = "ICOS Carbon Portal"
__license__     = "GPL-3.0"
__version__     = "0.0.2"
__maintainer__  = "ICOS Carbon Portal, elaborated products team"
__email__       = ['info@icos-cp.eu', 'karolina.pantazatou@nateko.lu.se']
__date__        = "2021-09-20"
__lastchange__  = ["Claudio DOnofrio"]
###############################################################################

#Import modules:
from datetime import date
import pandas as pd
import re
###############################################################################


#Function that returns a tuple with the start-date and
#end-date for a STILT station's available output:
def get_st_dates(st_dict):

    s_year = min(list(map(int, st_dict['years']))) #get start year
    e_year = max(list(map(int, st_dict['years']))) #get end year

    s_month = min(list(map(int, st_dict[str(s_year)]['months']))) #get start month
    e_month = max(list(map(int, st_dict[str(e_year)]['months']))) #get end month

    #Create start_date for STILT station output:
    stilt_s_date = date(s_year, s_month, 1)

    #Create end_date for STILT station output:
    stilt_e_date = date(e_year, e_month, 1)
    
    #Return tuple with STILT station start-date & end-date:
    return (stilt_s_date, stilt_e_date)
###############################################################################

def get_hours(hours):
    """
    STILT results are availabe in 3 hour times slots: 0 3 6 9 12 15 18 21
    You may choose to return only results for a specific time slot.
    Selection of hours: Input is expected to be a list or convertible to a list
    where each item can be casted to integer in the range of 0 to 23.
    
    If hours is empty or None, ALL Timeslots are returned.
    Valid results are returned as result with LOWER BOUND values.
    Example:    hours = [2,3,4] will return Timeslots for 0, 3 
                hours = [2,3,4,5,6] will return Timeslots for 0,3 and 6
                hours = [] return ALL
                hours = [10] returns timeslot 9
    """
    
    
    valid = [0,3,6,9,12,15,18,21,24]
    valid_hours = []
    for h in hours:
        # for back compatibility, we need to check input for str format
        # hh:mm -> convert to int
        if isinstance(h, str):
            if re.match('[0-9][0-9]:[0-9][0-9]', h):
                h = h[0:2]
                
        h = int(h)
        if h < 0 or h > 24:
            pass
        for i in range(0,8):
            if h >= valid[i] and h < valid[i+1]:
                valid_hours.append(valid[i])
                break                
    # make sure we have unique values
    valid_hours = list(set(valid_hours))
    
    if not hours:
        # return ALL in case no hours or only invalid hours are provided
        return valid[0:8]
    else:
        return valid_hours
    
###############################################################################

def parse(date):
    """
    convert date from different input formats:        

    Parameters
    ----------
    date :  STR 
            FLOAT (unix timestamp) 
            Python native datetime.date object
            pandas datetime object

    Returns
    -------
    datetime.date
    """
    out = False
    try:
        if isinstance(date, str):
            date = pd.to_datetime(date).date()
            out = True
        if isinstance(date, float) or isinstance(date, int):
            date = pd.to_datetime(date, unit='s').date()
            out = True
        
        if not out:            
            date = pd.to_datetime(date).date()
        
    except:
        date = None
    
    return date
    
    
def check_smonth(sdate, st_dict):
    """
    Function that checks if STILT model output is available for >= start date
    the check is for year and month only. 
    Example sdate = '2017-03-15' will return all stations which have data 
    for March 2017 or newer.
    """
    
    # if data exists in a year after sdate.year....True
    years = [int(y) for y in st_dict['years']]
    if any(y > sdate.year for y in years):
        return True 
    
    # if sdate.year is in years..check sdate.month 
    if sdate.year in years:
        months = [int(m) for m in st_dict[str(sdate.year)]['months']]
        if any(m >= sdate.month for m in months):
            return True
    return False


def check_emonth(edate, st_dict):
    """
    Function that checks if STILT model output is available for <= end date
    the check is for year and month only. 
    Example edate = '2017-03-15' will return all stations which have data 
    before or equal to March 2017.
    """
    
    # if data exists in a year before edate.year....True
    years = [int(y) for y in st_dict['years']]    
    if any(y < edate.year for y in years):
        return True 
    
    
    # if sdate.year is in years..check sdate.month 
    if edate.year in years:
        months = [int(m) for m in st_dict[str(edate.year)]['months']]
        if any(m <= edate.month for m in months):
            return True
    return False
    

def check_daterange(sdate, edate, st_dict):
    """
    Function that checks if STILT model output is available for
    >= sdate and <= edate
    the check is for year and month only. 
    Example daterange = ['2017-03-15','2017-05-28']' 
    will return all stations which have data >= March, 2017 and <= May 2017    
    """
    
    reflist = pd.date_range(start=sdate, end=edate, freq='MS').to_list()
    checklist = []            
    for y in st_dict['years']:
        for m in st_dict[y]['months']:
            checklist.append(pd.to_datetime(y + '-' + m))
    
    if list(set(checklist) & set(reflist)):
        return True
    



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    Created on Tue Aug 13 11:45:36 2019
    Check the usage of dataobjects through this
    It is a completely anonymous service to record how many times
    a dataobject is accessed.    
"""

__author__      = ["Claudio D'Onofrio"]
__credits__     = "ICOS Carbon Portal"
__license__     = "GPL-3.0"
__version__     = "0.1.0"
__maintainer__  = "ICOS Carbon Portal, elaborated products team"
__email__       = ['info@icos-cp.eu', 'claudio.donofrio@nateko.lu.se']
__status__      = "rc1"
__date__        = "2019-08-13"

import requests
import json

class Logfile():
    
    def __init__(self):
        self.server = 'https://restheart.icos-cp.eu/db/portaluse/'        
        self.flt = "?filter={'BinaryFileDownload':{'$exists':'true'}}&&count=true"        
        self._log = None        
        
        self.__getLogFile()
    
    @property
    def getLog(self):        
        return self._log                
    
    def getCount(self):        
        if self._log is not None:
            return self._log['_size']
        else:
            return 0

    def __getLogFile(self):
        url = self.server + self.flt
        r = requests.get(url)        
        self._log = json.loads(r.text)
        
if __name__ == '__main__':    
    print(str(Logfile().getCount()) + ' downloads counted')


#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    Description:      Class that creates objects to set and get the attributes
                      of a station for which STILT model output is available for.
"""

__credits__     = "ICOS Carbon Portal"
__license__     = "GPL-3.0"
__version__     = "0.1.0"
__maintainer__  = "ICOS Carbon Portal, elaborated products team"
__email__       = ['info@icos-cp.eu']
__status__      = "rc1"
__date__        = "2021-04-23"
#################################################################################

#Import modules
import os
import numpy as np
import pandas as pd
import requests
import json
import xarray as xr
import icoscp.const as CPC
from icoscp.stilt import timefuncs as tf
##############################################################################

class StiltStation():

    """
    Attributes: id:              STILT station ID (e.g. 'HTM150')
                locIdent:        String with latitude-longitude-altitude
                                 of STILT station (e.g. '35.34Nx025.67Ex00150')
                alt:             Station altitude (in meters above ground level)
                lat:             Station latitude
                lon:             Station longitude
                name:            STILT station long name
                icos:            Variable signifying if STILT station is an ICOS station
                years:           List of years for which STILT results are available for
                geoinfo:         Dictionary with nested dictionaries of geographical info

    """

    #Import modules:

    #Function that initializes the attributes of an object:
    def __init__(self, st_dict):

        #Object attributes:
        self._path_fp = CPC.STILTFP     # Path to location where STILT footprints are stored
        self._path_ts = CPC.STILTPATH   # Path to location where STILT time series are stored
        self._url = CPC.STILTTS         # URL to STILT information
        self.info = st_dict             # store the initial dict
        self.valid = False              # if input is dictionary, this will be True
        self.id = None                  # STILT ID for station (e.g. 'HTM150')
        self.lat = None
        self.lon = None
        self.alt = None
        self.locIdent = None
        self.name = None
        self.icos = None
        self.years = None
        self.geoinfo = None

        self._set(st_dict)

    #------------------------------------------------------------------------

    def _set(self, st_dict):
        #
        if all(item in st_dict.keys() for item in ['id', 'lat', 'lon', 'alt', 'locIdent', 'name', 'icos', 'years', 'geoinfo']):
            self.id = st_dict['id']
            self.lat = st_dict['lat']
            self.lon = st_dict['lon']
            self.alt = st_dict['alt']
            self.locIdent = st_dict['locIdent']
            self.name = st_dict['name']
            self.icos = st_dict['icos']
            self.years = sorted(st_dict['years'])
            self.geoinfo = st_dict['geoinfo']
            self.valid = True


    def __str__(self):
        # by default called when a an 'object' is printed

        out = {'id:': self.id,
               'name:': self.name,
               'lat:': self.lat,
               'lon:': self.lon,
               'alt [m]:': self.alt,
               'country': self.geoinfo['name']
               }
        return json.dumps(out)

    #----------------------------------------------------------------------------------------------------------
    def get_ts(self, start_date, end_date, hours=[], columns=''):
        """
        STILT concentration time series for a given time period,
        with optional selection of specific hours and columns.
        Returns time series in a pandas dataframe.

        Parameters
        ----------
        start_date : STR, FLOAT/INT (Unix timestamp), datetime object
            Example: start_date = '2018-01-01'
        end_date : STR, FLOAT/INT (Unix timestamp), datetime object
            Example: end_date = '2018-01-31'
        hours : STR | INT, optional
            If hours is empty or None, ALL Timeslots are returned.
            [0,3,6,9,12,15,18,21]
            
            Valid results are returned as result with LOWER BOUND values.
            For backwards compatibility, input for str format hh:mm is accepted
            Example:    hours = ["02:00",3,4] will return Timeslots for 0, 3
                        hours = [2,3,4,5,6] will return Timeslots for 0,3 and 6
                        hours = [] return ALL
                        hours = ["10", "10:00", 10] returns timeslot 9

        columns : TYPE, optional
            Valid entries are "default", "co2", "co", "rn", "wind", "latlon", "all"            
            default (or empty) will return
            ["isodate","co2.stilt","co2.fuel","co2.bio", "co2.background"]
            A full description of the 'columns' can be found at
            https://icos-carbon-portal.github.io/pylib/modules/#stilt

        Returns
        -------
        Pandas Dataframe
        """


        #Convert date-strings to date objs:
        s_date = tf.parse(start_date)
        e_date = tf.parse(end_date)
        hours = tf.get_hours(hours)

        # Check input parameters:
        if e_date < s_date:
            return False

        if not hours:
            return False

        #Create an empty dataframe to store the timeseries:
        df=pd.DataFrame({'A' : []})

        #Add headers:
        headers = {'Content-Type': 'application/json', 'Accept-Charset': 'UTF-8'}

        #Create an empty list, to store the new time range with available STILT model results:
        new_range=[]

        #Create a pandas dataframe containing one column of datetime objects with 3-hour intervals:
        #date_range = pd.date_range(start_date, end_date+dt.timedelta(hours=24), freq='3H')
        date_range = pd.date_range(s_date, e_date, freq='3H')

        #Loop through every Datetime object in the dataframe:
        for zDate in date_range:

            #Check if STILT results exist:
            if os.path.exists(self._path_fp + self.locIdent + '/' +
                              str(zDate.year)+'/'+str(zDate.month).zfill(2)+'/'+
                              str(zDate.year)+'x'+str(zDate.month).zfill(2)+'x'+str(zDate.day).zfill(2)+'x'+
                              str(zDate.hour).zfill(2)+'/'):

                #If STILT-results exist for the current Datetime object, append current Datetime object to list:
                new_range.append(zDate)

        #If the list is not empty:
        if len(new_range) > 0:

            #Assign the new time range to date_range:
            date_range = new_range

            #Get new starting date:
            fromDate = date_range[0].strftime('%Y-%m-%d')

            #Get new ending date:
            toDate = date_range[-1].strftime('%Y-%m-%d')

            #Store the STILT result column names to a variable:
            columns  = self.__columns(columns)

            #Store the STILT result data column names to a variable:
            data = '{"columns": '+columns+', "fromDate": "'+fromDate+'", "toDate": "'+toDate+'", "stationId": "'+self.id+'"}'

            #Send request to get STILT results:
            response = requests.post(self._url, headers=headers, data=data)

            #Check if response is successful:
            if response.status_code != 500:

                #Get response in json-format and read it in to a numpy array:
                output=np.asarray(response.json())

                #Convert numpy array with STILT results to a pandas dataframe:
                df = pd.DataFrame(output[:,:], columns=eval(columns))

                #Replace 'null'-values with numpy NaN-values:
                df = df.replace('null',np.NaN)

                #Set dataframe data type to float:
                df = df.astype(float)

                #Convert the data type of the 'date'-column to Datetime Object:
                df['date'] = pd.to_datetime(df['isodate'], unit='s')

                #Set 'date'-column as index:
                df.set_index(['date'],inplace=True)

                #Filter dataframe values by timeslots:
                hours = [str(h).zfill(2) for h in hours]
                df = df.loc[df.index.strftime('%H').isin(hours)]

            else:

                #Print message:
                print("\033[0;31;1m Error...\nToo big STILT dataset!\nSelect data for a shorter time period.\n\n")

        #Return dataframe:
        return df
    #----------------------------------------------------------------------------------------------------------

    def get_fp(self, start_date, end_date, hours=[]):
        """
        STILT footprints for a given time period,
        with optional selection of specific hours.
        Returns the footprints as xarray (http://xarray.pydata.org/en/stable/).
        with latitude, longitude, time, ppm per (micromol m-2 s-1).
        For more information about the STILT model at ICSO Carbon Portal
        please visit https://icos-carbon-portal.github.io/pylib/modules/#stilt

        Parameters
        ----------
        start_date : STR, FLOAT/INT (Unix timestamp), datetime object
            Example: start_date = '2018-01-01'
        end_date : STR, FLOAT/INT (Unix timestamp), datetime object
            Example: end_date = '2018-01-31'
        hours : STR | INT, optional
            If hours is empty or None, ALL Timeslots are returned.
            [0,3,6,9,12,15,18,21]
            
            Valid results are returned as result with LOWER BOUND values.
            For backwards compatibility, input for str format hh:mm is accepted
            Example:    hours = ["02:00",3,4] will return Timeslots for 0, 3
                        hours = [2,3,4,5,6] will return Timeslots for 0,3 and 6
                        hours = [] return ALL
                        hours = ["10", "10:00", 10] returns timeslot 9

        Returns
        -------
        Pandas Dataframe
        """

        #Convert date-strings to date objs:
        s_date = tf.parse(start_date)
        e_date = tf.parse(end_date)
        hours = tf.get_hours(hours)

        # Check input parameters:
        if e_date < s_date:
            return False

        if not hours:
            return False

        #Define & initialize footprint variable:
        fp = xr.DataArray()

        #Create a pandas dataframe containing one column of datetime objects with 3-hour intervals:
        date_range = pd.date_range(start_date, end_date, freq='3H')

        #Filter date_range by timeslots:
        date_range = [t for t in date_range if int(t.strftime('%H')) in hours]

        #Loop over all dates and store the corresponding fp filenames in a list:
        fp_files = [(self._path_fp + self.locIdent +'/'+
                     str(dd.year)+'/'+str(dd.month).zfill(2)+'/'+
                     str(dd.year)+'x'+str(dd.month).zfill(2)+'x'+
                     str(dd.day).zfill(2)+'x'+str(dd.hour).zfill(2)+'/foot')
                    for dd in date_range
                    if os.path.isfile(self._path_fp + self.locIdent +'/'+
                                      str(dd.year)+'/'+str(dd.month).zfill(2)+'/'+
                                      str(dd.year)+'x'+str(dd.month).zfill(2)+'x'+
                                      str(dd.day).zfill(2)+'x'+str(dd.hour).zfill(2)+'/foot')]

        #Concatenate xarrays on time axis:
        fp = xr.open_mfdataset(fp_files, concat_dim="time",
                               data_vars='minimal', coords='minimal',
                               compat='override', parallel=True)

        #Format time attributes:
        fp.time.attrs["standard_name"] = "time"
        fp.time.attrs["axis"] = "T"

        #Format latitude attributes:
        fp.lat.attrs["axis"] = "Y"
        fp.lat.attrs["standard_name"] = "latitude"

        #Format longitude attributes:
        fp.lon.attrs["axis"] = "X"
        fp.lon.attrs["standard_name"] = "longitude"

        #Return footprint array:
        return fp

    #Function that checks the selection of columns that are to be
    #returned with the STILT timeseries model output:
    def __columns(self, cols):

        # check for a valid entry. If not...return default
        valid = ["default", "co2", "co", "rn", "wind", "latlon", "all"]
        if cols not in valid:
            cols = 'default'

        #Check columns-input:
        if cols=='default':
            columns = ('["isodate","co2.stilt","co2.fuel","co2.bio", "co2.background"]')

        elif cols=='co2':
            columns = ('["isodate","co2.stilt","co2.fuel","co2.bio","co2.fuel.coal",'+
                    '"co2.fuel.oil","co2.fuel.gas","co2.fuel.bio","co2.energy",'+
                    '"co2.transport", "co2.industry","co2.others", "co2.cement",'+
                    '"co2.background"]')

        elif cols=='co':
            columns = ('["isodate", "co.stilt","co.fuel","co.bio","co.fuel.coal",'+
                    '"co.fuel.oil", "co.fuel.gas","co.fuel.bio","co.energy",'+
                    '"co.transport","co.industry", "co.others", "co.cement",'+
                    '"co.background"]')

        elif cols=='rn':
            columns = ('["isodate", "rn", "rn.era", "rn.noah"]')

        elif cols=='wind':
            columns = ('["isodate", "wind.dir", "wind.u", "wind.v"]')

        elif cols=='latlon':
            columns = ('["isodate", "latstart", "lonstart"]')

        elif cols=='all':
            columns = ('["isodate","co2.stilt","co2.fuel","co2.bio","co2.fuel.coal",'+
                    '"co2.fuel.oil","co2.fuel.gas","co2.fuel.bio","co2.energy",'+
                    '"co2.transport", "co2.industry","co2.others", "co2.cement",'+
                    '"co2.background", "co.stilt","co.fuel","co.bio","co.fuel.coal",'+
                    '"co.fuel.oil","co.fuel.gas","co.fuel.bio","co.energy",'+
                    '"co.transport","co.industry","co.others", "co.cement",'+
                    '"co.background","rn", "rn.era","rn.noah","wind.dir",'+
                    '"wind.u","wind.v","latstart","lonstart"]')


        #Return variable:
        return columns

# ----------------------------------- End of STILT Station Class ------------------------------------- #



"""
    The ICOS CarbonPortal package provides access to data hosted at the
    Carbon Portal. ( https://data.icos-cp.eu/ )    
    
    This library will load binary representations of digital data objects
    directly into memory. If you need store the data locally, we suggest
    you go to the website and "dowload" the data. The approach of this library
    is to free you from dowloadin and maintaining a local copy of data files.
    
    The ICOS Carbon Portal provides persistent digital object identifiers. Hence
    you only need to store a list of pid's and your script to reproduce
    always the same result, regardless on which computer you run it.
    (as long as you have an internet connection)    
"""

__credits__     = "ICOS Carbon Portal"
__license__     = ['Software: GPL-3.0','ICSO Data accessed with this library CC BY 4.0']
__version__     = "0.1.9"
__maintainer__  = "ICOS Carbon Portal, elaborated products team"
__email__       = ['info@icos-cp.eu', 'claudio.donofrio@nateko.lu.se']
__status__      = "stable"
__date__        = "2021-10-11"




#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""    
    Define global (to icoscp) constants
    mainly path variables and Url's used throught the package
    usage: import icoscp.const as CPC
    CPC.ST_PATH
    
"""
       
# Stilt specific

#local path to stiltstations if running on jupyter server
STILTPATH = '/data/stiltweb/stations/'

# Path to location where STILT footprints are stored
STILTFP = '/data/stiltweb/slots/'

# provide STILT station info from backend
STILTINFO = 'https://stilt.icos-cp.eu/viewer/stationinfo'

# online results
STILTTS = 'https://stilt.icos-cp.eu/viewer/stiltresult'